#ifndef NICVECTOR_CPP_UNIVERSITY_OF_OREGON_NIC
#define NICVECTOR_CPP_UNIVERSITY_OF_OREGON_NIC

#include "NICVector.h"
#include <algorithm>
#include <vector>
#include <iomanip>
#include <fstream>

//#ifdef __MANUAL_LAPACK //#######################################################
//#include <f2c.h>                                                              //
//#undef abs                                                                    //
//#endif  //--------------------------------------------------------------------//

#include "NICMatrix.h"

#include <cmath>

// explicit instantiations for NICVector types
template <class T> class NICVector;

//template class NICVector< nic_complex<double> >;
//template class NICVector< nic_complex<float> >;


template<class T>
NICVector<T>::NICVector() 
: m_size(0), m_data( 0 ) , m_noClobber( false )
{}

template<class T>
NICVector<T>::NICVector( int size ) 
: m_size( size ), m_data( new T[ size ] ), m_noClobber( false )
{
  for ( T* iter = m_data; iter < ( m_data + m_size ); ++iter ) 
  {
    *iter = 0; 
  }
}


template <class T>
NICVector<T>::NICVector( const NICVector<T>& vec ) 
: m_size( vec.m_size ), m_data( new T[m_size] ), m_noClobber( false )
{
  // maybe memcpy is faster?
  for ( int i = 0; i < m_size; ++i )
  {
    m_data[i] = vec.m_data[i];
  }
}

template <class T>
NICVector<T>::~NICVector()
{
  if ( 0 != m_data && !m_noClobber ) delete [] m_data;
}


template <class T>
void
NICVector<T>::zeroVector()
{
  constantVector( 0 );
}

template <class T>
void
NICVector<T>::constantVector( const T& v )
{
  for ( T* iter = m_data; iter < (m_data + m_size); ++iter )
  {
    *iter = v;
  }
}

template <class T>
void
NICVector<T>::randomVector( double min, double max )
{
  // TODO implement me!
}

template<class T>
void 
NICVector<T>::swapVectors( NICVector<T>& vec ) 
{
  std::swap( m_data, vec.m_data );
  std::swap( m_size, vec.m_size );
}

template <class T>
NICVector<T>&
NICVector<T>::operator=( const NICVector<T>& vec )
{
  // handle the memory management, only allocate
  // if we aren't big enough to hold vec
  if ( m_noClobber ) m_size = -1;
  if ( m_size < vec.m_size  )
  {
    delete [] m_data;
    m_data = new T[ vec.m_size ];
  }
  m_size = vec.m_size;

  for ( int i = 0; i < m_size; ++i )
  {
    m_data[i] = vec.m_data[i];
  }
  m_noClobber = false;
  return *this;
}

// for speed we have internal returns, normally this is bad style
template <class T>
bool
NICVector<T>::operator==( const NICVector<T>& vec ) const
{
  if ( m_size != vec.m_size ) return false;
  for ( int i = 0; i < m_size; ++i )
  {
    if ( m_data[i] != vec.m_data[i] ) return false;
  }
  return true;
}


// UPDATE THIS
// Casting everything to type T sucks. Should probably use RTTI to
// determine if complex and, if so, size *=2, cast to type of 
// complex and preceed
//
// a better solution is to use partial template specialization to
// handle complex values. adding this to the TODO list
template <class T>
void
NICVector<T>::normalize() 
{
  T norm = 0;


  for ( int i = 0; i < m_size; ++i )
  {
    norm += m_data[i] * m_data[i];
  }

  norm = std::sqrt(norm);

  for ( int i = 0; i < m_size; ++i) 
  {
    m_data[i] /= norm;
  }
}

template <>
void
NICVector<int>::normalize() 
{
  double norm = 0;


  for ( int i = 0; i < m_size; ++i )
  {
    norm += m_data[i] * m_data[i];
  }

  norm = std::sqrt(norm);

  for ( int i = 0; i < m_size; ++i) 
  {
    m_data[i]  = (int)((double)(m_data[i])/ norm);
  }
}

/****

The A is assumed to be normalized, 

v[i] = v[i] - A*v

****/
// UPDATE THIS
// Casting everything to type T sucks. Should probably use RTTI to
// determine if complex and, if so, size *=2, cast to type of 
// complex and preceed
// TODO update me
template <class T>
void
NICVector<T>::orthogonalize( NICMatrix<T>& mat ) 
{ 
  /*
  int i = 0, j = 0;
  double * temp = (double *)new T[n];
  double * a = (double *)(A->data);
  double * d = (double *)data;

  for (i = 0; i < n; i++) {
    temp[i] = d[i];
  }

  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      temp[i] -= a[i + n * j] * d[j];
    }
  }

   for (i = 0; i < n; i++) {
    d[i] = temp[i];
   }

   delete [] temp;
   */
}

template <class T>
T*
NICVector<T>::data()
{
  return m_data;
}

// we are setting the data to an external source. Since we no longer own the
// data, deleting can be catastrophic. Thus, we set the noclobber flag to
// prevent the deallocating unowned data
// Is this the best solution? I would say no.
//
// Why did I do this? I want to have a matrix of data. I pass a pointer for a
// row of data to the vector. Since the matrix owns the data, we want to make
// sure that we don't trample all over memory by trying to delete the vector
template <class T>
void
NICVector<T>::data( T* data, int size )
{
  if ( !m_noClobber )
  {
    delete [] m_data;
  }
  m_data = data;
  m_size = size;
  m_noClobber = true;
}

template <class T>
int
NICVector<T>::size() const
{
  return m_size;
}


template <class T>
std::ostream& operator<<( std::ostream& os, const NICVector<T>& vec )
{
  os << vec.m_size << " ";
  for ( int i = 0; i < vec.m_size; ++i )
  {
    os << vec.m_data[i] << " ";
  }
  return os;
}

template <class T>
std::istream& operator>> ( std::istream& is, NICVector<T>& vec )
{
  int size;
  is >> size;
  if ( vec.m_noClobber ) vec.m_size = -1;
  if ( size > vec.m_size ) 
  {
    delete [] vec.m_data;
    vec.m_data = new T[ size ];
  }
  vec.m_size = size;
  for ( int i = 0; i < vec.m_size; ++i )
  {
    is >> vec.m_data[i];
  }
  return is;
}

template <class T>
bool
NICVector<T>::read( std::istream& is, FileMode mode )
{
  // Assumptions: the input stream is binary
  // the data in the stream has the correct endian-ness for
  // the working machine

  // one might consider those assumptions to be bugs, and a more
  // complete implementation would address them
  // for now, we just want this to work

  // another note: this is likely not the most efficient way to do
  // this. obvious optimizations are welcome ;-)

  // what size is the data? create a temp buffer to read values
  size_t sizeOfT = sizeof( T );
  char* nextValue = new char[ sizeOfT ];
  T* nextValueT;

  std::vector< T > buffer;
  int bufferSize = 0;

  // * try to read the next value (with read)
  // * if successful (the ios::good call checks for that)
  // * cast the pointer to the target type
  // * then push the value onto the back of the buffer
  while ( ( is.read( nextValue, sizeOfT ) ).good() )
  {
    nextValueT = (T*)(nextValue);
    buffer.push_back( *nextValueT );
    ++bufferSize;
  }

  if ( m_noClobber ) m_size = -1;
  if ( m_size < bufferSize )
  {
    delete [] m_data;
    m_data = new T[ bufferSize ];
  }

  m_size = bufferSize;
  for ( int i = 0; i < bufferSize; ++i )
  {
    m_data[i] = buffer[i];
  }

  // another shortcoming, we want to return false on failure.
  // TODO this should be the first thing fixed about this method
  std::cout << "read " << bufferSize << " values" << std::endl;
  
  delete [] nextValue;
  return true;
}

template <class T>
bool
NICVector<T>::read( std::istream& is, int numberToRead, FileMode mode )
{
  size_t sizeOfT = sizeof( T );
  char* nextValue = new char[ sizeOfT ];
  T* nextValueT;

  if ( m_noClobber ) 
  {
    m_size = -1;
    m_noClobber = false;
  }
  else
  {
      delete [] m_data;
  }

  if ( m_size < numberToRead )
  {
    m_data = new T[ numberToRead ];
  }
  
  m_size = numberToRead;
  int count = 0;

  // * try to read the next value (with read)
  // * if successful (the ios::good call checks for that)
  // * cast the pointer to the target type
  // * then push the value onto the back of the buffer
  while ( (count < numberToRead ) &&
          ( is.read( nextValue, sizeOfT ) ).good() )
  {
    nextValueT = (T*)(nextValue);
    m_data[count] = *nextValueT;
    ++count;
  }

  delete [] nextValue;

  return ( count == numberToRead );
}

template <class T>
bool
NICVector<T>::write( std::ostream& os, FileMode mode )
{
  return write( os, m_size, mode );
}

template <class T>
bool
NICVector<T>::write( std::ostream& os, int numberToWrite, FileMode mode )
{
  size_t sizeOfT = sizeof(T);
  char* nextValue;

  for ( int i = 0; i < numberToWrite; ++i )
  {
    nextValue = (char*)( m_data + i );
    os.write( nextValue, sizeOfT );
  }
  // TODO we could add more sophisticated error handling here
  // an alternative would be to change the api to match the
  // read/write api provided by the standard library
  return true;
}

template <class T>
bool NICVector<T>::checkFileBits( FileMode mode )
{
    // check for binary multiple set
    if ( (NICVector<T>::BINARY | NICVector<T>::ASCII) & mode )
    {
        return false;
    }

    if ( (NICVector<T>::LITTLEENDIAN | NICVector<T>::BIGENDIAN) & mode )
    {
        return false;
    }

    if ( (NICVector<T>::LITTLEENDIAN | NICVector<T>::NATIVE) & mode )
    {
        return false;
    }

    if ( (NICVector<T>::NATIVE | NICVector<T>::BIGENDIAN) & mode )
    {
        return false;
    }

    return true;
}


// instantiate the stream operators
template 
std::ostream& operator<< <>( std::ostream& out, const NICVector<double>& v );
template 
std::ostream& operator<< <>( std::ostream& out, const NICVector<float>& v );

template class NICVector<double>;
template class NICVector<float>;
template class NICVector<int>;

#endif
// NICVECTOR_CPP_UNIVERSITY_OF_OREGON_NIC
